@@ -1,30 +1,31 @@
 /*
- * filefuncs.c - Builtin functions that provide initial minimal iterface
+ * filefuncs.c - Builtin functions that provide initial minimal interface
  *		 to the file system.
  *
  * Arnold Robbins, update for 3.1, Mon Nov 23 12:53:39 EST 1998
  * Arnold Robbins and John Haque, update for 3.1.4, applied Mon Jun 14 13:55:30 IDT 2004
  * Arnold Robbins and Andrew Schorr, revised for new extension API, May 2012.
  * Arnold Robbins, add fts(), August 2012
+ * Arnold Robbins, add statvfs(), November 2015
  */
 
 /*
- * Copyright (C) 2001, 2004, 2005, 2010, 2011, 2012, 2013, 2014
+ * Copyright (C) 2001, 2004, 2005, 2010-2021, 2023,
  * the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -71,17 +72,22 @@
 #include <string.h>
 #include <unistd.h>
 
-#include <sys/types.h>
-#include <sys/stat.h>
-
 #ifdef HAVE_SYS_PARAM_H
 #include <sys/param.h>
 #endif /* HAVE_SYS_PARAM_H */
 
-#ifdef MAJOR_IN_MKDEV
-#include <sys/mkdev.h>
-#elif defined(MAJOR_IN_SYSMACROS)
+#if HAVE_SYS_SYSMACROS_H
 #include <sys/sysmacros.h>
+#elif HAVE_SYS_MKDEV_H
+#include <sys/mkdev.h>
+#endif /* HAVE_SYS_MKDEV_H */
+
+#include <sys/types.h>
+
+#include <sys/stat.h>
+
+#if defined(HAVE_SYS_STATVFS_H) && defined(HAVE_STATVFS)
+#include <sys/statvfs.h>
 #endif
 
 #include "gawkapi.h"
@@ -120,13 +126,16 @@ static long long
 get_inode(const char *fname)
 {
 	HANDLE fh;
+	BOOL ok;
 	BY_HANDLE_FILE_INFORMATION info;
 
 	fh = CreateFile(fname, 0, 0, NULL, OPEN_EXISTING,
 			FILE_FLAG_BACKUP_SEMANTICS, NULL);
 	if (fh == INVALID_HANDLE_VALUE)
 		return 0;
-	if (GetFileInformationByHandle(fh, &info)) {
+	ok = GetFileInformationByHandle(fh, &info);
+	CloseHandle(fh);
+	if (ok) {
 		long long inode = info.nFileIndexHigh;
 
 		inode <<= 32;
@@ -138,26 +147,23 @@ get_inode(const char *fname)
 #endif
 
 static const gawk_api_t *api;	/* for convenience macros to work */
-static awk_ext_id_t *ext_id;
+static awk_ext_id_t ext_id;
 static awk_bool_t init_filefuncs(void);
 static awk_bool_t (*init_func)(void) = init_filefuncs;
 static const char *ext_version = "filefuncs extension: version 1.0";
 
 int plugin_is_GPL_compatible;
 
-/*  do_chdir --- provide dynamically loaded chdir() builtin for gawk */
+/*  do_chdir --- provide dynamically loaded chdir() function for gawk */
 
 static awk_value_t *
-do_chdir(int nargs, awk_value_t *result)
+do_chdir(int nargs, awk_value_t *result, struct awk_ext_func *unused)
 {
 	awk_value_t newdir;
 	int ret = -1;
 
 	assert(result != NULL);
 
-	if (do_lint && nargs != 1)
-		lintwarn(ext_id, _("chdir: called with incorrect number of arguments, expecting 1"));
-
 	if (get_argument(0, AWK_STRING, & newdir)) {
 		ret = chdir(newdir.str_value.str);
 		if (ret < 0)
@@ -400,7 +406,8 @@ fill_stat_array(const char *name, awk_array_t array, struct stat *sbuf)
 	array_set_numeric(array, "gid", sbuf->st_gid);
 	array_set_numeric(array, "size", sbuf->st_size);
 #ifdef __MINGW32__
-	array_set_numeric(array, "blocks", (sbuf->st_size + 4095) / 4096);
+	array_set_numeric(array, "blocks", (double)((sbuf->st_size +
+		device_blocksize() - 1) / device_blocksize()));
 #else
 	array_set_numeric(array, "blocks", sbuf->st_blocks);
 #endif
@@ -448,7 +455,7 @@ fill_stat_array(const char *name, awk_array_t array, struct stat *sbuf)
 		}
 	}
 
-	array_set(array, "type", make_const_string(type, strlen(type), &tmp));
+	array_set(array, "type", make_const_string(type, strlen(type), & tmp));
 
 	return 0;
 }
@@ -456,7 +463,7 @@ fill_stat_array(const char *name, awk_array_t array, struct stat *sbuf)
 /* do_stat --- provide a stat() function for gawk */
 
 static awk_value_t *
-do_stat(int nargs, awk_value_t *result)
+do_stat(int nargs, awk_value_t *result, struct awk_ext_func *unused)
 {
 	awk_value_t file_param, array_param;
 	char *name;
@@ -467,19 +474,17 @@ do_stat(int nargs, awk_value_t *result)
 
 	assert(result != NULL);
 
-	if (nargs != 2 && nargs != 3) {
-		if (do_lint)
-			lintwarn(ext_id, _("stat: called with wrong number of arguments"));
+	/* file is first arg, array to hold results is second */
+	if (! get_argument(0, AWK_STRING, & file_param)) {
+		warning(ext_id, _("stat: first argument is not a string"));
 		return make_number(-1, result);
 	}
 
-	/* file is first arg, array to hold results is second */
-	if (   ! get_argument(0, AWK_STRING, & file_param)
-	    || ! get_argument(1, AWK_ARRAY, & array_param)) {
-		warning(ext_id, _("stat: bad parameters"));
+	if (! get_argument(1, AWK_ARRAY, & array_param)) {
+		warning(ext_id, _("stat: second argument is not an array"));
 		return make_number(-1, result);
 	}
-	
+
 	if (nargs == 3) {
 		statfunc = stat;
 	}
@@ -490,7 +495,7 @@ do_stat(int nargs, awk_value_t *result)
 	/* always empty out the array */
 	clear_array(array);
 
-	/* stat the file, if error, set ERRNO and return */
+	/* stat the file; if error, set ERRNO and return */
 	ret = statfunc(name, & sbuf);
 	if (ret < 0) {
 		update_ERRNO_int(errno);
@@ -502,6 +507,60 @@ do_stat(int nargs, awk_value_t *result)
 	return make_number(ret, result);
 }
 
+#if defined(HAVE_SYS_STATVFS_H) && defined(HAVE_STATVFS)
+
+/* do_statvfs --- provide a statvfs() function for gawk */
+
+static awk_value_t *
+do_statvfs(int nargs, awk_value_t *result, struct awk_ext_func *unused)
+{
+	awk_value_t file_param, array_param;
+	char *name;
+	awk_array_t array;
+	int ret;
+	struct statvfs vfsbuf;
+
+	assert(result != NULL);
+
+	/* file is first arg, array to hold results is second */
+	if (   ! get_argument(0, AWK_STRING, & file_param)
+	    || ! get_argument(1, AWK_ARRAY, & array_param)) {
+		warning(ext_id, _("stat: bad parameters"));
+		return make_number(-1, result);
+	}
+
+	name = file_param.str_value.str;
+	array = array_param.array_cookie;
+
+	/* always empty out the array */
+	clear_array(array);
+
+	/* statvfs the filesystem; if error, set ERRNO and return */
+	ret = statvfs(name, & vfsbuf);
+	if (ret < 0) {
+		update_ERRNO_int(errno);
+		return make_number(ret, result);
+	}
+
+	array_set_numeric(array, "bsize", vfsbuf.f_bsize);	/* filesystem block size */
+	array_set_numeric(array, "frsize", vfsbuf.f_frsize);	/* fragment size */
+	array_set_numeric(array, "blocks", vfsbuf.f_blocks);	/* size of fs in f_frsize units */
+	array_set_numeric(array, "bfree", vfsbuf.f_bfree);	/* # free blocks */
+	array_set_numeric(array, "bavail", vfsbuf.f_bavail);	/* # free blocks for unprivileged users */
+	array_set_numeric(array, "files", vfsbuf.f_files);	/* # inodes */
+	array_set_numeric(array, "ffree", vfsbuf.f_ffree);	/* # free inodes */
+	array_set_numeric(array, "favail", vfsbuf.f_favail);	/* # free inodes for unprivileged users */
+#ifndef _AIX
+	array_set_numeric(array, "fsid", vfsbuf.f_fsid);	/* filesystem ID */
+#endif
+	array_set_numeric(array, "flag", vfsbuf.f_flag);	/* mount flags */
+	array_set_numeric(array, "namemax", vfsbuf.f_namemax);	/* maximum filename length */
+
+
+	return make_number(ret, result);
+}
+#endif
+
 /* init_filefuncs --- initialization routine */
 
 static awk_bool_t
@@ -513,6 +572,7 @@ init_filefuncs(void)
 
 #ifndef __MINGW32__
 	/* at least right now, only FTS needs initializing */
+#define FTS_NON_RECURSIVE	FTS_STOP	/* Don't step into directories.  */
 	static struct flagtab {
 		const char *name;
 		int value;
@@ -524,6 +584,7 @@ init_filefuncs(void)
 		ENTRY(FTS_PHYSICAL),
 		ENTRY(FTS_SEEDOT),
 		ENTRY(FTS_XDEV),
+		{"FTS_SKIP", FTS_NON_RECURSIVE},
 		{ NULL, 0 }
 	};
 
@@ -540,7 +601,7 @@ init_filefuncs(void)
 }
 
 #ifdef __MINGW32__
-/*  do_fts --- walk a heirarchy and fill in an array */
+/*  do_fts --- walk a hierarchy and fill in an array */
 
 /*
  * Usage from awk:
@@ -549,7 +610,7 @@ init_filefuncs(void)
  */
 
 static awk_value_t *
-do_fts(int nargs, awk_value_t *result)
+do_fts(int nargs, awk_value_t *result, struct awk_ext_func *unused)
 {
 	fatal(ext_id, _("fts is not supported on this system"));
 
@@ -570,7 +631,7 @@ fill_stat_element(awk_array_t element_array, const char *name, struct stat *sbuf
 
 	stat_array = create_array();
 	if (stat_array == NULL) {
-		warning(ext_id, _("fill_stat_element: could not create array"));
+		warning(ext_id, _("fill_stat_element: could not create array, out of memory"));
 		fts_errors++;
 		return;
 	}
@@ -636,10 +697,10 @@ fill_default_elements(awk_array_t element_array, const FTSENT *const fentry, awk
 	}
 }
 
-/* process --- process the heirarchy */
+/* process --- process the hierarchy */
 
 static void
-process(FTS *heirarchy, awk_array_t destarray, int seedot)
+process(FTS *hierarchy, awk_array_t destarray, int seedot, int skipset)
 {
 	FTSENT *fentry;
 	awk_value_t index, value;
@@ -648,13 +709,18 @@ process(FTS *heirarchy, awk_array_t destarray, int seedot)
 
 	/* path is full path,  pathlen is length thereof */
 	/* name is name in directory, namelen is length thereof */
-	while ((fentry = fts_read(heirarchy)) != NULL) {
+	while ((fentry = fts_read(hierarchy)) != NULL) {
 		bad_ret = awk_false;
 
 		switch (fentry->fts_info) {
 		case FTS_D:
 			/* directory */
+
+			if (skipset && fentry->fts_level == 0)
+				fts_set(hierarchy, fentry, FTS_SKIP);
+
 			/* create array to hold entries */
+			/* this will be empty if doing FTS_SKIP */
 			newdir_array = create_array();
 			if (newdir_array == NULL) {
 				warning(ext_id, _("fts-process: could not create array"));
@@ -755,7 +821,7 @@ process(FTS *heirarchy, awk_array_t destarray, int seedot)
 	}
 }
 
-/*  do_fts --- walk a heirarchy and fill in an array */
+/*  do_fts --- walk a hierarchy and fill in an array */
 
 /*
  * Usage from awk:
@@ -764,39 +830,39 @@ process(FTS *heirarchy, awk_array_t destarray, int seedot)
  */
 
 static awk_value_t *
-do_fts(int nargs, awk_value_t *result)
+do_fts(int nargs, awk_value_t *result, struct awk_ext_func *unused)
 {
 	awk_value_t pathlist, flagval, dest;
 	awk_flat_array_t *path_array = NULL;
 	char **pathvector = NULL;
-	FTS *heirarchy;
+	FTS *hierarchy;
 	int flags;
 	size_t i, count;
 	int ret = -1;
 	static const int mask = (
 		  FTS_COMFOLLOW | FTS_LOGICAL | FTS_NOCHDIR | FTS_PHYSICAL
-		| FTS_SEEDOT | FTS_XDEV);
+		| FTS_SEEDOT | FTS_XDEV | FTS_NON_RECURSIVE);
 
 	assert(result != NULL);
 	fts_errors = 0;		/* ensure a fresh start */
 
-	if (do_lint && nargs != 3)
+	if (nargs > 3)
 		lintwarn(ext_id, _("fts: called with incorrect number of arguments, expecting 3"));
 
 	if (! get_argument(0, AWK_ARRAY, & pathlist)) {
-		warning(ext_id, _("fts: bad first parameter"));
+		warning(ext_id, _("fts: first argument is not an array"));
 		update_ERRNO_int(EINVAL);
 		goto out;
 	}
 
 	if (! get_argument(1, AWK_NUMBER, & flagval)) {
-		warning(ext_id, _("fts: bad second parameter"));
+		warning(ext_id, _("fts: second argument is not a number"));
 		update_ERRNO_int(EINVAL);
 		goto out;
 	}
 
 	if (! get_argument(2, AWK_ARRAY, & dest)) {
-		warning(ext_id, _("fts: bad third parameter"));
+		warning(ext_id, _("fts: third argument is not an array"));
 		update_ERRNO_int(EINVAL);
 		goto out;
 	}
@@ -825,10 +891,12 @@ do_fts(int nargs, awk_value_t *result)
 	}
 	flags &= mask;	/* turn off anything else */
 
+	if (flags & FTS_NON_RECURSIVE)
+		flags |= FTS_NOCHDIR;
+
 	/* make pathvector */
 	count = path_array->count + 1;
-	emalloc(pathvector, char **, count * sizeof(char *), "do_fts");
-	memset(pathvector, 0, count * sizeof(char *));
+	ezalloc(pathvector, char **, count * sizeof(char *), "do_fts");
 
 	/* fill it in */
 	count--;	/* ignore final NULL at end of vector */
@@ -837,15 +905,15 @@ do_fts(int nargs, awk_value_t *result)
 
 
 	/* clear dest array */
-	if (! clear_array(dest.array_cookie)) {
-		warning(ext_id, _("fts: clear_array() failed\n"));
-		goto out;
-	}
+	assert(clear_array(dest.array_cookie));
 
 	/* let's do it! */
-	if ((heirarchy = fts_open(pathvector, flags, NULL)) != NULL) {
-		process(heirarchy, dest.array_cookie, (flags & FTS_SEEDOT) != 0);
-		fts_close(heirarchy);
+	hierarchy = fts_open(pathvector, flags & ~FTS_NON_RECURSIVE, NULL);
+	if (hierarchy != NULL) {
+		process(hierarchy, dest.array_cookie,
+			(flags & FTS_SEEDOT) != 0,
+			(flags & FTS_NON_RECURSIVE) != 0);
+		fts_close(hierarchy);
 
 		if (fts_errors == 0)
 			ret = 0;
@@ -863,10 +931,13 @@ out:
 #endif	/* ! __MINGW32__ */
 
 static awk_ext_func_t func_table[] = {
-	{ "chdir",	do_chdir, 1 },
-	{ "stat",	do_stat, 2 },
+	{ "chdir",	do_chdir, 1, 1, awk_false, NULL },
+	{ "stat",	do_stat, 3, 2, awk_false, NULL },
 #ifndef __MINGW32__
-	{ "fts",	do_fts, 3 },
+	{ "fts",	do_fts, 3, 3, awk_false, NULL },
+#endif
+#if defined(HAVE_SYS_STATVFS_H) && defined(HAVE_STATVFS)
+	{ "statvfs",	do_statvfs, 2, 2, awk_false, NULL },
 #endif
 };
 
