@@ -2,10 +2,10 @@
  * interpret.h ---  run a list of instructions.
  */
 
-/* 
- * Copyright (C) 1986, 1988, 1989, 1991-2022,
+/*
+ * Copyright (C) 1986, 1988, 1989, 1991-2023,
  * the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
  *
@@ -212,32 +212,42 @@ top:
 
 			case Node_var_new:
 uninitialized_scalar:
-				if (op != Op_push_arg_untyped) {
-					/* convert untyped to scalar */
-					m->type = Node_var;
-					m->var_value = dupnode(Nnull_string);
-				}
 				if (do_lint)
 					lintwarn(isparam ?
 						_("reference to uninitialized argument `%s'") :
 						_("reference to uninitialized variable `%s'"),
 								save_symbol->vname);
-				if (op != Op_push_arg_untyped)
+
+				if (op != Op_push_arg_untyped) {
+					// convert very original untyped to scalar
+					m->type = Node_var;
+					m->var_value = dupnode(Nnull_string);
+
+					// set up local param by value
 					m = dupnode(Nnull_string);
+				}
+
 				UPREF(m);
 				PUSH(m);
 				break;
 
 			case Node_elem_new:
-				if (op != Op_push_arg_untyped) {
-					/* convert untyped to scalar */
-					m = elem_new_to_scalar(m);
-				}
 				if (do_lint)
 					lintwarn(isparam ?
 						_("reference to uninitialized argument `%s'") :
 						_("reference to uninitialized variable `%s'"),
 								save_symbol->vname);
+
+				if (op != Op_push_arg_untyped) {
+					// convert very original untyped to scalar
+					m->type = Node_var;
+					m->var_value = dupnode(Nnull_string);
+
+					// set up local param by value
+					DEREF(m);
+					m = dupnode(Nnull_string);
+				}
+
 				PUSH(m);
 				break;
 
@@ -265,8 +275,7 @@ uninitialized_scalar:
 				PUSH(m);
 		 		break;
 			}
- 			/* else
-				fall through */
+ 			/* fall through */
 		case Op_push_array:
 			PUSH(pc->memory);
 			break;
@@ -334,7 +343,9 @@ uninitialized_scalar:
 				}
 			}
 
-			if (r->type == Node_val || r->type == Node_elem_new)
+			if (r->type == Node_val
+			    || r->type == Node_var
+			    || r->type == Node_elem_new)
 				UPREF(r);
 			PUSH(r);
 			break;
@@ -576,6 +587,7 @@ uninitialized_scalar:
 plus:
 			t1 = TOP_NUMBER();
 			r = make_number(t1->numbr + x2);
+			r->numbr = fix_nan_sign(t1->numbr, x2, r->numbr);
 			DEREF(t1);
 			REPLACE(r);
 			break;
@@ -590,6 +602,7 @@ plus:
 minus:
 			t1 = TOP_NUMBER();
 			r = make_number(t1->numbr - x2);
+			r->numbr = fix_nan_sign(t1->numbr, x2, r->numbr);
 			DEREF(t1);
 			REPLACE(r);
 			break;
@@ -780,6 +793,7 @@ mod:
 			break;
 
 		case Op_store_field:
+		case Op_store_field_exp:
 		{
 			/* field assignment optimization,
 			 * see awkgram.y (optimize_assignment)
@@ -802,6 +816,10 @@ mod:
 			UNFIELD(*lhs, r);
 			/* field variables need the string representation: */
 			force_string(*lhs);
+			if (op == Op_store_field_exp) {
+				UPREF(*lhs);
+				PUSH(*lhs);
+			}
 		}
 			break;
 
