@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1999-2021 the Free Software Foundation, Inc.
+ * Copyright (C) 1999-2023 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -119,17 +119,12 @@ set_prof_file(const char *file)
 void
 init_profiling_signals()
 {
-#ifdef __DJGPP__
-	signal(SIGINT, dump_and_exit);
-	signal(SIGQUIT, just_dump);
-#else  /* !__DJGPP__ */
 #ifdef SIGHUP
 	signal(SIGHUP, dump_and_exit);
 #endif
 #ifdef SIGUSR1
 	signal(SIGUSR1, just_dump);
 #endif
-#endif /* !__DJGPP__ */
 }
 
 /* indent --- print out enough tabs */
@@ -373,8 +368,7 @@ pprint(INSTRUCTION *startp, INSTRUCTION *endp, int flags)
 				break;
 
 			default:
-				fprintf(stderr, "Got unexpected type %s\n", nodetype2str(m->type));
-				cant_happen();
+				cant_happen("got unexpected type %s", nodetype2str(m->type));
 			}
 
 			switch (pc->opcode) {
@@ -545,15 +539,28 @@ cleanup:
 			break;
 
 		case Op_store_field:
+		case Op_store_field_exp:
+		{
+			char *assignment, *final;
+
 			t1 = pp_pop(); /* field num */
 			if (is_binary(t1->type))
 				pp_parenthesize(t1);
 			t2 = pp_pop(); /* r.h.s. */
-			fprintf(prof_fp, "$%s%s%s", t1->pp_str, op2str(pc->opcode), t2->pp_str);
+			assignment = pp_group3(t1->pp_str, op2str(pc->opcode), t2->pp_str);
+			final = pp_group3("$", assignment, "");
+			efree(assignment);
 			pp_free(t2);
 			pp_free(t1);
-			if ((flags & IN_FOR_HEADER) == 0)
-				pc = end_line(pc);
+			if (pc->opcode == Op_store_field_exp)
+				pp_push(pc->opcode, final, CAN_FREE, NULL);
+			else {
+				fprintf(prof_fp, "%s", final);
+				efree(final);
+				if ((flags & IN_FOR_HEADER) == 0)
+					pc = end_line(pc);
+			}
+		}
 			break;
 
 		case Op_concat:
@@ -581,7 +588,7 @@ cleanup:
 
 		case Op_K_delete_loop:
 			/* Efficency hack not in effect because of exec_count instruction */
-			cant_happen();
+			cant_happen("unexpected opcode %s", opcode2str(pc->opcode));
 			break;
 
 		case Op_in_array:
@@ -669,17 +676,17 @@ cleanup:
 		case Op_K_print_rec:
 			if (pc->opcode == Op_K_print_rec)
 				// instead of `print $0', just `print'
-				tmp = strdup("");
+				tmp = estrdup("", 0);
 			else if (pc->redir_type != 0) {
 				// Avoid turning printf("hello\n") into printf(("hello\n"))
 				NODE *n = pp_top();
 
 				if (pc->expr_count == 1
-				    && n->pp_str[0] == '(' 
+				    && n->pp_str[0] == '('
 				    && n->pp_str[n->pp_len - 1] == ')') {
 					n = pp_pop();
 
-					tmp = strdup(n->pp_str);
+					tmp = estrdup(n->pp_str, strlen(n->pp_str));
 					pp_free(n);
 				} else
 					tmp = pp_list(pc->expr_count, "()", ", ");
@@ -705,8 +712,7 @@ cleanup:
 		case Op_push_re:
 			if (pc->memory->type != Node_regex && (pc->memory->flags & REGEX) == 0)
 				break;
-			/* else
-				fall through */
+			/* fall through */
 		case Op_match_rec:
 		{
 			if (pc->memory->type == Node_regex) {
@@ -999,9 +1005,11 @@ cleanup:
 			fprintf(prof_fp, "%s (", op2str(pc->opcode));
 			pprint(pc->nexti, ip1->switch_start, NO_PPRINT_FLAGS);
 			t1 = pp_pop();
-			fprintf(prof_fp, "%s) {\n", t1->pp_str);
+			fprintf(prof_fp, "%s) {", t1->pp_str);
 			if (pc->comment)
 				print_comment(pc->comment, 0);
+			else
+				fprintf(prof_fp, "\n");
 			pp_free(t1);
 			pprint(ip1->switch_start, ip1->switch_end, NO_PPRINT_FLAGS);
 			indent(SPACEOVER);
@@ -1023,9 +1031,11 @@ cleanup:
 
 			indent_in();
 			if (pc->comment != NULL) {
-				if (pc->comment->memory->comment_type == EOL_COMMENT)
+				if (pc->comment->memory->comment_type == EOL_COMMENT) {
 					fprintf(prof_fp, "\t%s", pc->comment->memory->stptr);
-				else {
+					if (pc->comment->comment != NULL)
+						print_comment(pc->comment->comment, indent_level);
+				} else {
 					fprintf(prof_fp, "\n");
 					print_comment(pc->comment, indent_level);
 				}
@@ -1219,7 +1229,7 @@ cleanup:
 			break;
 
 		default:
-			cant_happen();
+			cant_happen("unexpected opcode %s", opcode2str(pc->opcode));
 		}
 
 		if (pc == endp)
@@ -1788,7 +1798,7 @@ pp_list(int nargs, const char *paren, const char *delim)
 			len += r->pp_len + delimlen;
 			if (r->pp_comment != NULL) {
 				comment = (INSTRUCTION *) r->pp_comment;
-				len += comment->memory->stlen + indent_level + 1;	// comment\n ident
+				len += comment->memory->stlen + indent_level + 1;	// comment\n indent
 			}
 		}
 		if (paren != NULL) {
@@ -1992,7 +2002,7 @@ pp_func(INSTRUCTION *pc, void *data ATTRIBUTE_UNUSED)
 		print_comment(pc->comment, -1);	/* -1 ==> don't indent */
 
 	indent(pc->nexti->exec_count);
-	
+
 	bool malloced = false;
 	char *name = adjust_namespace(func->vname, & malloced);
 	fprintf(prof_fp, "%s %s(", op2str(Op_K_function), name);
